/* fb.c - Copyright (C) Tomas Berndtsson, tomas@nocrew.org
 *
 * Functions for handling the frame buffer in Linux/m68k.
 *
 * Uses oFBis, a frame buffer lib made by Martin Mitchell
 *             with small changes by Tomas Berndtsson.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fb/libfb.h>
#include "fb.h"

static FB *fb;

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

static void signal_handler(int sig)
{
  fprintf(stderr,"fb: signal %d caught and exits program!\n", sig);
  fb_exit();
  exit(1);
}

static void setup_signals()
{
  sigset_t mask;
  struct sigaction sigs;

  sigemptyset(&mask);

  sigs.sa_handler = signal_handler;
  sigs.sa_mask = mask;
  sigs.sa_flags = SA_ONESHOT;
  sigs.sa_restorer = NULL;

  sigaction(SIGHUP, &sigs, NULL);
  sigaction(SIGINT, &sigs, NULL);
  sigaction(SIGQUIT, &sigs, NULL);
  sigaction(SIGILL, &sigs, NULL);
  sigaction(SIGTRAP, &sigs, NULL);
  sigaction(SIGBUS, &sigs, NULL);
  sigaction(SIGFPE, &sigs, NULL);
  sigaction(SIGSEGV, &sigs, NULL);
  sigaction(SIGTERM, &sigs, NULL);
}

int fb_setup(int bpp_required)
{
  fb = FBopen(NULL, FB_OPEN_NEW_VC);
  if(fb == NULL)
    return 1;

  if(fb->vinf.bits_per_pixel < bpp_required) {
    FBclose(fb);
    return 1;
  }

  setup_signals();

  return 0;
}

void fb_exit()
{
  if(fb != NULL)
    FBclose(fb);
}

FB *fb_getFB()
{
  return fb;
}

void fb_setcolor(int col, int red, int green, int blue)
{
  FBCMAP *fbcmap;

  fbcmap = FBgetcmap(fb);

  if(fbcmap == NULL)
    return;

  col = max(0,min(col,(1<<fb->vinf.bits_per_pixel)-1));

  fbcmap->red[col]   = red;
  fbcmap->green[col] = green;
  fbcmap->blue[col]  = blue;

  fbcmap->start = fbcmap->end = col;
  fbcmap->len = 1;

  FBputcmap(fb, fbcmap);
  FBfreecmap(fbcmap);
}

void fb_setpalette(int *red, int *green, int *blue)
{
  int nr_cols,col;
  FBCMAP *fbcmap;

  fbcmap = FBgetcmap(fb);

  if(fbcmap == NULL)
    return;

  nr_cols = 1<<fb->vinf.bits_per_pixel;

  for(col=0 ; col<nr_cols ; col++) {
    fbcmap->red[col]   = red[col];
    fbcmap->green[col] = green[col];
    fbcmap->blue[col]  = blue[col];
  }

  fbcmap->start = 0;
  fbcmap->end = nr_cols-1;
  fbcmap->len = nr_cols;

  FBputcmap(fb, fbcmap);
  FBfreecmap(fbcmap);
}

void fb_go_tty()
{
  FBVTswitchoriginal(fb);
}

void fb_go_fb()
{
  FBVTswitchfb(fb);
  sleep(1); /* This is an ugly hack! */
}

void fb_waitkey()
{
  FBgetchar(fb);
}

void fb_putpixel(int x, int y, int col)
{
  col = max(0,min(col,(1<<fb->vinf.bits_per_pixel)-1));

  FBputpixel(fb, x, y, col);
}

void fb_clear_area(int x, int y, int w, int h)
{
  FBBLTPBLK fbb;
  int i;
  unsigned short dummy;

  fbb.b_wd = w;
  fbb.b_ht = h;
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x00000000;
  fbb.s_xmin = 0;
  fbb.s_ymin = 0;
  fbb.s_form = &dummy;
  fbb.s_nxwd = 0;
  fbb.s_nxln = 0;
  fbb.s_nxpl = 0;
  fbb.d_xmin = x;
  fbb.d_ymin = y;
  fbb.d_form = fb->sbuf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb);
}


/* Stuff used for selecting an area on the frame buffer */

typedef struct
{
  unsigned short *buf;
  int x,y,w,h;
} SavedBlock;

static SavedBlock *save_back(int x, int y, int w, int h,
			     int min_x, int max_x, int min_y, int max_y)
{
  int i, size, words;
  FBBLTPBLK fbb;
  SavedBlock *block;
  
  block = (SavedBlock *)malloc(sizeof(SavedBlock));
  if(block == NULL)
    return NULL;

  block->x = max(min_x,x);
  block->y = max(min_y,y);
  block->w = min(max_x-block->x+1,w);
  block->h = min(max_y-block->y+1,h);

  words = (block->w%16 ? block->w/16+1 : block->w/16);
  size = words*block->h*fb->vinf.bits_per_pixel*2;
  block->buf = (unsigned short *)malloc(size);
  if(block->buf == NULL) {
    free(block);
    return NULL;
  }

  fbb.b_wd = block->w;
  fbb.b_ht = block->h;
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x03030303;
  fbb.s_xmin = block->x;
  fbb.s_ymin = block->y;
  fbb.s_form = fb->sbuf;
  fbb.s_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.s_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.s_nxpl = 2;
  fbb.d_xmin = 0;
  fbb.d_ymin = 0;
  fbb.d_form = block->buf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = words*fbb.d_nxwd;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb);

  return block;
}

static void restore_back(SavedBlock *block)
{
  int i, words;
  FBBLTPBLK fbb;

  words = (block->w%16 ? block->w/16+1 : block->w/16);
  
  fbb.b_wd = block->w;
  fbb.b_ht = block->h;
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x03030303;
  fbb.s_xmin = 0;
  fbb.s_ymin = 0;
  fbb.s_form = block->buf;
  fbb.s_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.s_nxln = words*fbb.s_nxwd;
  fbb.s_nxpl = 2;
  fbb.d_xmin = block->x;
  fbb.d_ymin = block->y;
  fbb.d_form = fb->sbuf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb);

  free(block->buf);
  free(block);
}

static void restore_back_keep(SavedBlock *block)
{
  int i, words;
  FBBLTPBLK fbb;

  words = (block->w%16 ? block->w/16+1 : block->w/16);
  
  fbb.b_wd = block->w;
  fbb.b_ht = block->h;
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x03030303;
  fbb.s_xmin = 0;
  fbb.s_ymin = 0;
  fbb.s_form = block->buf;
  fbb.s_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.s_nxln = words*fbb.s_nxwd;
  fbb.s_nxpl = 2;
  fbb.d_xmin = block->x;
  fbb.d_ymin = block->y;
  fbb.d_form = fb->sbuf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb);
}

static void set_cursor(int x, int y,
		       int min_x, int max_x, int min_y, int max_y)
{
  int i;
  FBBLTPBLK fbb;
  unsigned short curs_cross_gfx[] = { 0x1000,0x1000,0x1000,0xfe00,
				      0x1000,0x1000,0x1000 };
  unsigned short curs_cross_mask[] = { 0x2800,0x2800,0xee00,0x0000,
				       0xee00,0x2800,0x2800 };
  x-=3;
  y-=3;

  if(x < min_x) {
    fbb.b_wd = 7-(min_x-x);
    fbb.s_xmin = min_x-x;
    fbb.d_xmin = 0;
  } else {
    fbb.b_wd = min(max_x-x+1,7);
    fbb.s_xmin = 0;
    fbb.d_xmin = min(x,max_x);
  }
  if(y < min_y) {
    fbb.b_ht = 7-(min_y-y);
    fbb.s_ymin = min_y-y;
    fbb.d_ymin = 0;
  } else {
    fbb.b_ht = min(max_y-y+1,7);
    fbb.s_ymin = 0;
    fbb.d_ymin = min(y,max_y);
  }
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.s_nxwd = 2;
  fbb.s_nxln = 2;
  fbb.s_nxpl = 0;
  fbb.d_form = fb->sbuf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  /* Mask */
  fbb.s_form = curs_cross_mask;
  fbb.op_tab = 0x04040404; /* and inverted */
  FBbitblt(fb, &fbb);  

  /* Cross */
  fbb.s_form = curs_cross_gfx;
  fbb.op_tab = 0x07070707; /* or */
  FBbitblt(fb, &fbb);  
}

static void set_hdotline(int x, int y, int w)
{
  FBBLTPBLK fbb;
  int i;
  unsigned short dot_line = 0xaaaa; /* dotted */

  fbb.b_wd = w;
  fbb.b_ht = 1;
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x03030303;
  fbb.s_xmin = 0;
  fbb.s_ymin = 0;
  fbb.s_form = &dot_line;
  fbb.s_nxwd = 0;
  fbb.s_nxln = 0;
  fbb.s_nxpl = 0;
  fbb.d_xmin = x;
  fbb.d_ymin = y;
  fbb.d_form = fb->sbuf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb);
}

static void set_vdotline(int x, int y, int h)
{
  FBBLTPBLK fbb;
  int i, line_len;
  unsigned short dummy;

  line_len = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;

  fbb.b_wd = 1;
  fbb.b_ht = max(1,h/2);
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x0f0f0f0f;
  fbb.s_xmin = 0;
  fbb.s_ymin = 0;
  fbb.s_form = &dummy;
  fbb.s_nxwd = 0;
  fbb.s_nxln = 0;
  fbb.s_nxpl = 0;
  fbb.d_xmin = x;
  fbb.d_ymin = y/2;
  fbb.d_form = fb->sbuf + (y%2 ? line_len/2 : 0);
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = line_len * 2;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb); /* white pixels */

  fbb.b_ht = (y%2 ? fbb.b_ht-1 : fbb.b_ht);
  fbb.op_tab = 0x00000000;
  fbb.d_form = fb->sbuf + (y%2 ? line_len : line_len/2);

  if(fbb.b_ht > 1)
    FBbitblt(fb, &fbb); /* black pixels */
}

#define CORNER_BOX_SIZE 5

static void set_corner_box(int x, int y)
{
  FBBLTPBLK fbb;
  int i;
  unsigned short dot_box[] = { 0xaaaa, 0x5555, 0xaaaa, 0x5555,
			       0xaaaa, 0x5555, 0xaaaa, 0x5555,
			       0xaaaa, 0x5555, 0xaaaa, 0x5555,
			       0xaaaa, 0x5555, 0xaaaa, 0x5555 };
  
  fbb.b_wd = fbb.b_ht = CORNER_BOX_SIZE;
  fbb.plane_ct = fb->vinf.bits_per_pixel;
  fbb.fg_col = fbb.bg_col = 0;
  fbb.op_tab = 0x03030303;
  fbb.s_xmin = 0;
  fbb.s_ymin = 0;
  fbb.s_form = dot_box;
  fbb.s_nxwd = 0;
  fbb.s_nxln = 2;
  fbb.s_nxpl = 0;
  fbb.d_xmin = x;
  fbb.d_ymin = y;
  fbb.d_form = fb->sbuf;
  fbb.d_nxwd = fb->vinf.bits_per_pixel*2;
  fbb.d_nxln = fb->vinf.xres_virtual*fb->vinf.bits_per_pixel/8;
  fbb.d_nxpl = 2;
  fbb.p_addr = NULL;
  fbb.p_nxln = fbb.p_nxpl = fbb.p_mask = 0;
  for(i=0; i<12; i++)
    fbb.space[i] = 0;

  FBbitblt(fb, &fbb);
}

static void set_select_box(int x1, int y1, int x2, int y2)
{
  set_vdotline(x1, y1, y2-y1+1);
  set_vdotline(x2, y1, y2-y1+1);
  set_hdotline(x1, y1, x2-x1+1);
  set_hdotline(x1, y2, x2-x1+1);

  if(x2-x1 >= CORNER_BOX_SIZE && y2-y1 >= CORNER_BOX_SIZE) {
    set_corner_box(x1+1, y1+1);
    set_corner_box(x2-CORNER_BOX_SIZE, y1+1);
    set_corner_box(x1+1, y2-CORNER_BOX_SIZE);
    set_corner_box(x2-CORNER_BOX_SIZE, y2-CORNER_BOX_SIZE);
  }
}


/* This code can probably be cleaned up some day */

/* Returns 1 if an area is selected, 0 if not (== if key pressed). */
int fb_select_area(SelectionArea *info)
{
  FBEVENT ev;
  int quit, ud, rect, dr, first;
  int abs_x,abs_y, new_x,new_y;
  int stick_x,stick_y;
  SavedBlock *cursor, *left,*right,*up,*down;

  if(info->x1 >= 0) {
    info->x1 = max(info->min_x,min(info->max_x,info->x1));
    info->y1 = max(info->min_y,min(info->max_y,info->y1));
    info->x2 = max(info->min_x,min(info->max_x,info->x2));
    info->y2 = max(info->min_y,min(info->max_y,info->y2));

    left  = save_back(info->x1,info->y1, 
		      CORNER_BOX_SIZE+1,info->y2-info->y1+1,
		      info->min_x,info->max_x,info->min_y,info->max_y);
    right = save_back(max(info->x1,info->x2-CORNER_BOX_SIZE),info->y1, 
		      CORNER_BOX_SIZE+1,info->y2-info->y1+1,
		      info->min_x,info->max_x,info->min_y,info->max_y);
    up    = save_back(info->x1+CORNER_BOX_SIZE,info->y1, 
		      max(1,info->x2-info->x1+1-CORNER_BOX_SIZE),1,
		      info->min_x,info->max_x,info->min_y,info->max_y);
    down  = save_back(info->x1+CORNER_BOX_SIZE,info->y2, 
		      max(1,info->x2-info->x1+1-CORNER_BOX_SIZE),1,
		      info->min_x,info->max_x,info->min_y,info->max_y);
    if(left==NULL || right==NULL || up==NULL || down==NULL) {
      fprintf(stderr, "l: %p r: %p u: %p d: %p\n",left,right,up,down);
      fb_exit();
      exit(1);
    }

    set_select_box(info->x1, info->y1, info->x2, info->y2);

    abs_x = (info->x2+info->x1) / 2;
    abs_y = (info->y2+info->y1) / 2;

    rect = 1;
    first = 0;

  } else {
    abs_x = (info->max_x+info->min_x) / 2;
    abs_y = (info->max_y+info->min_y) / 2;

    rect = 0;
    first = 1;
  }


  /* save background */
  cursor = save_back(abs_x-3,abs_y-3, 7,7,
		     info->min_x, info->max_x, info->min_y, info->max_y);
  if(cursor == NULL) { 
    fprintf(stderr,"cursor==NULL\n");
    fb_exit();
    exit(1);
  }

  set_cursor(abs_x, abs_y, info->min_x, info->max_x, info->min_y, info->max_y);

  quit = 0;
  dr = 0;
  do {
    FBgetevent(fb, &ev);

    ud=0;
    if(ev.type == FBMouseEvent)
      {	
	new_x = abs_x + (ev.mouse.x>>1);
	new_y = abs_y + (ev.mouse.y>>1);

	if(!dr && !rect) { /* if mouse button is not down already */
	  if(!(ev.mouse.buttons & 0x4)) { /* and is being pressed */
	    dr=1;
	    stick_x = abs_x;
	    stick_y = abs_y;
	  } else {
	    new_x = min(info->max_x,max(new_x,info->min_x));
	    new_y = min(info->max_y,max(new_y,info->min_y));

	    if(abs_x != new_x || abs_y != new_y) {
	      abs_x = new_x;
	      abs_y = new_y;
	      ud = 1;
	    }
	  }
	} else if(dr && !rect) { /* if mouse button is down */
	  if(ev.mouse.buttons & 0x4) { /* and is being released */
	    switch(dr) {
	    case 1:
	      info->x1 = stick_x; info->y1 = stick_y;
	      info->x2 = abs_x;   info->y2 = abs_y;
	      break;
	    case 2:
	      info->x1 = abs_x;   info->y1 = stick_y;
	      info->x2 = stick_x; info->y2 = abs_y;
	      break;
	    case 3:
	      info->x1 = stick_x; info->y1 = abs_y;
	      info->x2 = abs_x;   info->y2 = stick_y;
	      break;
	    case 4:
	      info->x1 = abs_x;   info->y1 = abs_y;
	      info->x2 = stick_x; info->y2 = stick_y;
	      break;
	    }
	    dr=0;
	    rect=1;

	  } else {
	    switch(dr) {
	    case 1: /* lower right, default */
	      new_x = min(info->max_x,max(new_x,stick_x));
	      new_y = min(info->max_y,max(new_y,stick_y));
	      break;
	    case 2: /* lower left */
	      new_x = max(info->min_x,min(new_x,stick_x));
	      new_y = min(info->max_y,max(new_y,stick_y));
	      break;
	    case 3: /* upper right */
	      new_x = min(info->max_x,max(new_x,stick_x));
	      new_y = max(info->min_y,min(new_y,stick_y));
	      break;
	    case 4: /* upper left */
	      new_x = max(info->min_x,min(new_x,stick_x));
	      new_y = max(info->min_y,min(new_y,stick_y));
	      break;
	    }

	    if(abs_x != new_x || abs_y != new_y) {
	      abs_x = new_x;
	      abs_y = new_y;
	      ud = 1;
	    }
	  }
	} else if(rect) {
	  if(!(ev.mouse.buttons & 0x4)) {
	    if(abs_x >= info->x1 && abs_x <= min(info->x1+CORNER_BOX_SIZE,info->x2) &&
	       abs_y >= info->y1 && abs_y <= min(info->y1+CORNER_BOX_SIZE,info->y2)) {
	      /* upper left corner */
	      dr=4;
	      rect=0;
	      stick_x = info->x2;
	      stick_y = info->y2;
	      abs_x = info->x1;
	      abs_y = info->y1;
	    } else if(abs_x <= info->x2 && abs_x >= max(info->x2-CORNER_BOX_SIZE,info->x1) &&
		      abs_y >= info->y1 && abs_y <= min(info->y1+CORNER_BOX_SIZE,info->y2)) {
	      /* upper right corner */
	      dr=3;
	      rect=0;
	      stick_x = info->x1;
	      stick_y = info->y2;
	      abs_x = info->x2;
	      abs_y = info->y1;
	    } else if(abs_x >= info->x1 && abs_x <= min(info->x1+CORNER_BOX_SIZE,info->x2) &&
		      abs_y <= info->y2 && abs_y >= max(info->y2-CORNER_BOX_SIZE,info->y1)) {
	      /* lower left corner */
	      dr=2;
	      rect=0;
	      stick_x = info->x2;
	      stick_y = info->y1;
	      abs_x = info->x1;
	      abs_y = info->y2;
	    } else if(abs_x <= info->x2 && abs_x >= max(info->x2-CORNER_BOX_SIZE,info->x1) &&
		      abs_y <= info->y2 && abs_y >= max(info->y2-CORNER_BOX_SIZE,info->y1)) {
	      /* lower right corner */
	      dr=1;
	      rect=0;
	      stick_x = info->x1;
	      stick_y = info->y1;
	      abs_x = info->x2;
	      abs_y = info->y2;
	    }
	  } else {
	    new_x = min(info->max_x,max(new_x,info->min_x));
	    new_y = min(info->max_y,max(new_y,info->min_y));

	    if(abs_x != new_x || abs_y != new_y) {
	      abs_x = new_x;
	      abs_y = new_y;
	      ud = 1;
	    }
	  }

        }

      } else if(ev.type == FBKeyEvent) {
	static int hidden=0;

	/* only care if key is pressed, not released!
	 * only care about keys with an ascii code too, 
	 * otherwise VT switching doesn't seem to work. */
	if(!(ev.key.keycode & 0x80) && ev.key.ascii) {
	  info->key = ev.key.ascii;
	  quit = 1;
	} else if(!hidden && ev.key.state & 0x03) { /* left or right shift */
	  if(!first) {
	    restore_back_keep(left);
	    restore_back_keep(right);
	    restore_back_keep(up);
	    restore_back_keep(down);
	    hidden=1;
	  }
	} else if(hidden && !(ev.key.state & 0x03)) {
	  set_select_box(info->x1,info->y1,info->x2,info->y2);
	  hidden=0;
	}
      } else {
	/* Unknown event type, just ignore */
	fprintf(stderr,"fb_select_area(): Unknown event type %d ignored\n",ev.type);
      }

    if(ud)
      {
	/* restore old backgrounds */
	restore_back(cursor);
	
	if(dr && !first) {
	  restore_back(left);
	  restore_back(right);
	  restore_back(up);
	  restore_back(down);
	} else if(dr)
	  first=0;

	/* save new backgrounds */
	switch(dr) {
	case 1:
	  left  = save_back(stick_x,stick_y, 
			    CORNER_BOX_SIZE+1,abs_y-stick_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  right = save_back(max(stick_x,abs_x-CORNER_BOX_SIZE),stick_y, 
			    CORNER_BOX_SIZE+1,abs_y-stick_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  up    = save_back(stick_x+CORNER_BOX_SIZE,stick_y, 
			    max(1,abs_x-stick_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  down  = save_back(stick_x+CORNER_BOX_SIZE,abs_y, 
			    max(1,abs_x-stick_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  if(left==NULL || right==NULL || up==NULL || down==NULL) {
	    fprintf(stderr, "l: %p r: %p u: %p d: %p\n",left,right,up,down);
	    fb_exit();
	    exit(1);
	  }

	  set_select_box(stick_x, stick_y, abs_x, abs_y);
	  break;

	case 2:
	  left  = save_back(abs_x,stick_y, 
			    CORNER_BOX_SIZE+1,abs_y-stick_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  right = save_back(max(abs_x,stick_x-CORNER_BOX_SIZE),stick_y, 
			    CORNER_BOX_SIZE+1,abs_y-stick_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  up    = save_back(abs_x+CORNER_BOX_SIZE,stick_y, 
			    max(1,stick_x-abs_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  down  = save_back(abs_x+CORNER_BOX_SIZE,abs_y, 
			    max(1,stick_x-abs_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  if(left==NULL || right==NULL || up==NULL || down==NULL) {
	    fprintf(stderr, "l: %p r: %p u: %p d: %p\n",left,right,up,down);
	    fb_exit();
	    exit(1);
	  }

	  set_select_box(abs_x, stick_y, stick_x, abs_y);
	  break;

	case 3:
	  left  = save_back(stick_x,abs_y, 
			    CORNER_BOX_SIZE+1,stick_y-abs_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  right = save_back(max(stick_x,abs_x-CORNER_BOX_SIZE),abs_y, 
			    CORNER_BOX_SIZE+1,stick_y-abs_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  up    = save_back(stick_x+CORNER_BOX_SIZE,abs_y, 
			    max(1,abs_x-stick_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  down  = save_back(stick_x+CORNER_BOX_SIZE,stick_y, 
			    max(1,abs_x-stick_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  if(left==NULL || right==NULL || up==NULL || down==NULL) {
	    fprintf(stderr, "l: %p r: %p u: %p d: %p\n",left,right,up,down);
	    fb_exit();
	    exit(1);
	  }

	  set_select_box(stick_x, abs_y, abs_x, stick_y);
	  break;

	case 4:
	  left  = save_back(abs_x,abs_y, 
			    CORNER_BOX_SIZE+1,stick_y-abs_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  right = save_back(max(abs_x,stick_x-CORNER_BOX_SIZE),abs_y, 
			    CORNER_BOX_SIZE+1,stick_y-abs_y+1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  up    = save_back(abs_x+CORNER_BOX_SIZE,abs_y, 
			    max(1,stick_x-abs_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  down  = save_back(abs_x+CORNER_BOX_SIZE,stick_y, 
			    max(1,stick_x-abs_x+1-CORNER_BOX_SIZE),1,
			    info->min_x,info->max_x,info->min_y,info->max_y);
	  if(left==NULL || right==NULL || up==NULL || down==NULL) {
	    fprintf(stderr, "l: %p r: %p u: %p d: %p\n",left,right,up,down);
	    fb_exit();
	    exit(1);
	  }

	  set_select_box(abs_x, abs_y, stick_x, stick_y);
	  break;
	}

	/* save background */
	cursor = save_back(abs_x-3,abs_y-3, 7,7, 
			   info->min_x,info->max_x,info->min_y,info->max_y);
	if(cursor == NULL) {
	  fprintf(stderr,"cursor==NULL\n");
	  fb_exit();
	  exit(1);
	}

	set_cursor(abs_x, abs_y, info->min_x, info->max_x, info->min_y, info->max_y);
      }

  } while(!quit);

  /* restore background */
  restore_back(cursor);
  if(!first) {
    restore_back(left);
    restore_back(right);
    restore_back(up);
    restore_back(down);
  }

  if(rect)
    return 1;
  else
    return 0;
}
