/* hp.c - Copyright (C) Tomas Berndtsson, tomas@nocrew.org
 *
 * Functions for controlling HP Scanjet scsi scanners.
 *
 * Based on hpscanpbm, written by David Etherton, etherton@netcom.com
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include <scsi/sg.h>

#include "hp.h"
#include "pnm.h"

#define min(a,b) ( (a) < (b) ? (a) : (b) )

extern int verbose;
extern char *infotext;

static int sg_xfer(int fid,unsigned char *src,int len,unsigned char *dest,
		   int reply_len)
{
        int count, result;
        static struct sg_header hdr;
        static char buf[4096 + sizeof(struct sg_header)];
        hdr.pack_len = sizeof(hdr) + len;
        hdr.reply_len = sizeof(hdr) + reply_len;
        memcpy(buf, &hdr, sizeof(hdr));
        memcpy(buf + sizeof(hdr), src, len);
        if (write(fid, buf, hdr.pack_len) == -1) {
                perror("sg_xfer(write)");
                return -1;
        }
        if ((count = read(fid,buf,hdr.reply_len)) == -1) {
                perror("sg_xfer(read)");
                return -1;
        }
        result = ((struct sg_header*)buf)->result;
        if (result) {
                fprintf(stderr,"sg_xfer: result is %d\n",result);
        }
        if (count > sizeof(struct sg_header)) {
                count -= sizeof(struct sg_header);
                memcpy(dest,buf + sizeof(hdr),count);
        }
        else
                count = 0;
        return count;
}

static int inquire_device(int fid,char *mfg,char *model)
{
        static unsigned char inquiry[] = { 0x12,0,0,0,255,0 };
        unsigned char reply[256];
        if (sg_xfer(fid,inquiry,sizeof(inquiry),reply,sizeof(reply)) < 16)
                return -1;
        memcpy(mfg,reply + 8, 8);
        memcpy(model,reply + 16, 16);
        return reply[0];
}

static int write_data(int fid,char *data,int len)
{
        char write_6[4096];
        memset(write_6,0,6);
        write_6[0] = 0xa;
        write_6[2] = len >> 16;
        write_6[3] = len >> 8;
        write_6[4] = len;
        memcpy(write_6 + 6, data, len);
        if (sg_xfer(fid,write_6,6 + len,0,0) < 0)
                return -1;
        return 0;
}

static int read_data(int fid,char *dest,int len)
{
        static unsigned char read_6[6] = { 8,0,0,0,0,0 };
        read_6[2] = len >> 16;
        read_6[3] = len >> 8;
        read_6[4] = len;
        if ((len=sg_xfer(fid,read_6,6,dest,len)) < 0)
                return -1;
        /* dump(dest,len); */
        return len;
}

static int write_string(int fid,char *format,...)
{
	int res;

        va_list args;
        char buf[1024];
        va_start(args, format);
        vsprintf(buf,format,args);
        res = write_data(fid,buf,strlen(buf));
	if(res == -1)
	  fprintf(stderr,"write_string(%s): Couldn't be written!\n",buf);

	return res;
}

static int inquire(int fid,char *inq)
{
        char reply[128];
        write_string(fid,inq);
        if (read_data(fid,reply,sizeof(reply))) {
                /* dump(reply,128); */
                /* this should be less haphazard */
                return atoi(strchr(reply,'d') + 1);
        }
        else
                return -1;
}

static int norm8(double value)
{
	if (value < -1.0)
		value = -1.0;
	else if (value > 1.0)
		value = 1.0;
	return (int)(value * 127.0);
}

static PNMPic *setup_scan(int fid, ScanSettings *ss)
{
        int mode, pixels_per_line, bytes_per_line, lines;
	PNMPic *dest;

	/*	dump_settings(ss); */

	dest = (PNMPic *)malloc(sizeof(PNMPic));
	if(dest == NULL)
	  return NULL;

        /* reset scanner; returns all paramters to defaults */
        write_string(fid,"\033E");

        /* set resolution, in DPI */
        write_string(fid,"\033*a%dR",ss->xdpi);
        write_string(fid,"\033*a%dS",ss->ydpi);

        /* set scan extents, in 1/300'ths of an inch */
        write_string(fid,"\033*f%dX",ss->x);
        write_string(fid,"\033*f%dY",ss->y);
        write_string(fid,"\033*f%dP",ss->width);
        write_string(fid,"\033*f%dQ",ss->height);

        /* (the original scanjet only allows -1, 0, or 1) */
        write_string(fid,"\033*a%dL",norm8(ss->brightness));
        write_string(fid,"\033*a%dK",norm8(ss->contrast));

        /* greyscale and color modes seems to need to have their
                outputs inverted; thresholded and dithered modes
                do not (the default) */
        if (ss->format > FORMAT_DITHERED)
                write_string(fid,"\033*a1I");
        mode = 4;
        /* data format is thresholded by default; pixel format
                is 8 pixels/byte */
        if (ss->format == FORMAT_DITHERED) {
                /* data format: dithered */
                write_string(fid,"\033*a3T");
        }
        else if (ss->format == FORMAT_GREYSCALE) {
                /* data format is greyscale, pixel format is
                        one pixel/byte */
                write_string(fid,"\033*a4T");
                write_string(fid,"\033*a8G");
                mode = 5;
	}
        else if (ss->format == FORMAT_TRUECOLOR) {
                /* data format is 24bit color, pixel format
                        is one pixel/three bytes */
                write_string(fid,"\033*a5T");
                write_string(fid,"\033*a24G");
                mode = 6;
        }

        /* inquire resulting size of image after setting it up */
        pixels_per_line = inquire(fid,"\033*s1024E");
        bytes_per_line = inquire(fid,"\033*s1025E");
        lines = inquire(fid,"\033*s1026E");

        if (verbose) {
	  fprintf(stderr,"mode %d, ",mode);
	  fprintf(stderr,"%d pixels per line, %d bytes, %d lines high, ",
		  pixels_per_line, bytes_per_line, lines);
	  fprintf(stderr,"xdpi %d, ydpi %d\n",ss->xdpi,ss->ydpi);
	}

	if(mode > 4)
	  dest->max_color = 255;
	else
	  dest->max_color = 1;

	dest->mode = mode;
	dest->xres = pixels_per_line;
	dest->yres = lines;
	dest->length = bytes_per_line * lines;

	return dest;
}

PNMPic *hp_scan_mem(int fid,ScanSettings* ss)
{
        int len, left;
	char *buf;
	PNMPic *dest;

	dest = setup_scan(fid, ss);
	if(dest == NULL)
	  return NULL;

	dest->buf = (char *)malloc(dest->length);
	if(dest->buf == NULL)
	  return NULL;

        write_string(fid,"\033*f0S");           /* begin scan! */

        /* consume all data */
	buf = dest->buf;
	left = dest->length;
        while (left && (len = read_data(fid,buf,min(BUFFER_SIZE,left)))) {
	  if(len < 0) {
	    fprintf(stderr,"Something went wrong while reading!\n");
	    break;
	  } else {
	    left -= len;
	    buf += len;
	  }
        }

	if(left) {
	  free(dest->buf);
	  free(dest);
	  return NULL;
	}

	return dest;
}

int hp_scan_file(int outfd, int fid,ScanSettings* ss)
{
        int len, left;
	char *buf;
	char tmp[1024],pretext[1024];
	PNMPic *dest;

	dest = setup_scan(fid, ss);
	if(dest == NULL)
	  return -1;

	sprintf(tmp,
		"P%d\n"
		"# %s\n"
		"%d %d\n",
		dest->mode, infotext, dest->xres, dest->yres);
	if(dest->mode > 4)
	  sprintf(pretext, "%s%d\n", tmp, dest->max_color);
	else
	  strcpy(pretext, tmp);

	if(write(outfd, pretext, strlen(pretext)) == -1) {
	  perror("writing pretext");
	  free(dest);
	  return -1;
	}

	buf = (char *)malloc(BUFFER_SIZE);
	if(buf == NULL)
	  return -1;

        write_string(fid,"\033*f0S");           /* begin scan! */

        /* consume all data */
	left = dest->length;
        while (left && (len = read_data(fid,buf,min(BUFFER_SIZE,left)))) {
	  if(len < 0) {
	    fprintf(stderr,"Something went wrong while reading!\n");
	    break;
	  } else {
	    left -= len;
	    if(write(outfd, buf, len) == -1) {
	      perror("writing output file");
	      break;
	    }
	  }
        }

	free(buf);
	free(dest);

	if(left)
	  return -1;
	
	return 0;
}


int hp_is_scanner(char *devname)
{
        int fd = open(devname, O_RDWR);
        if (fd != -1) {
                char mfr[8], model[16];
                int type = inquire_device(fd,mfr,model);
                if (type == 3 && !strncmp(mfr,"HP",2))
                        return fd;
                close(fd);
        }
        return -1;
}

int hp_find_scanner(void)
{
        char buf[16];
        int i, fd;
        for (i='a'; i <= 'z'; i++) {
                sprintf(buf,"/dev/sg%c",i);
                if ((fd = hp_is_scanner(buf)) != -1) {
                        if (verbose)
                                fprintf(stderr,
                                        "HP Scanjet found on '%s'\n",buf);
                        return fd;
                }
        }
        return -1;
}

