/* hpscanfb.c - Copyright (C) Tomas Berndtsson, tomas@nocrew.org
 * 
 * Main program for hpscanfb.
 * Vaguely based on code from hpscanpbm by David Etherton, etherton@netcom.com
 *                                     and Steve Enns, ennss@nhrisv.nhrc.sk.doe.ca
 *
 * This program is subject to the terms and conditions of the GNU General
 * Public License version 2 or, at own choice, any later version. See the 
 * file COPYING in the main directory of this archive for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "version.h"
#include "hp.h"
#include "pnm.h"
#include "fb.h"

int verbose=0, no_fb=0, auto_prescan=0, prescan_format=FORMAT_GREYSCALE;
char *outfile=NULL;
char *infotext="Scanned with hpscanfb, Copyright (C) Tomas Berndtsson (tomas@nocrew.org)";

int scan_to_file(int scanfd, ScanSettings *ss);
int parse_options(int argc, char *argv[], ScanSettings *ss);


/* these should really be inline functions in hp.c */
#define in2scanner(val) ((val) * 300.0)
#define cm2scanner(val) (((val) * 300.0) / 2.54)
#define mm2scanner(val) (((val) * 300.0) / 25.4)
#define pixels2scanner(val,dpi) (((val) * 300.0) / (dpi))
#define scanner2in(val) ((val) / 300.0)
#define scanner2cm(val) (((val) * 2.54) / 300.0)
#define scanner2mm(val) (((val) * 25.4) / 300.0)
#define scanner2pixels(val,dpi) (((val) * (dpi)) / 300.0)

int main(int argc,char *argv[])
{
  int scanfd;
  int quit,prescan;
  PNMPic *scan=NULL;
  SelectionArea selarea;

  /* default is 50dpi binarized A4 page */
  /* This gives a picture of about 414x585 pixels */
  ScanSettings ss = { 50, 50, 0, 0, mm2scanner(210), mm2scanner(297), FORMAT_TRUECOLOR, 0.0, 0.0 };
  ScanSettings fbss; /* Used for scanning to the frame buffer */
  ScanSettings set; /* after parse_options(), used to reset the settings */

  if(parse_options(argc, argv, &ss))
    exit(1);

  set = ss;

  scanfd = hp_find_scanner();
  if(scanfd == -1) {
    fprintf(stderr,"No HP Scanjet found.\n");
    exit(2);
  }

  if(!no_fb) {
    if(fb_setup(1)) {
      fprintf(stderr,"%s: Could not open frame buffer!\n",argv[0]);
      exit(1);
    }

    
    quit = 0;
    prescan = 1;
    while(!quit) {
      if(prescan) {
	fbss = ss;
	fbss.format = prescan_format;
	/* calculate some other stuff here later, e.g. dpi */

	if(scan != NULL) {
	  fb_clear_area(0, 0, scan->xres, scan->yres);
	  pnm_close(scan);
	}	

	scan = hp_scan_mem(scanfd,&fbss);
	if(scan == NULL) {
	  close(scanfd);
	  exit(1);
	}

	pnm_draw(fb_getFB(), scan);

	selarea.min_x = selarea.min_y = 0;
	selarea.max_x = scan->xres-1;
	selarea.max_y = scan->yres-1;
	selarea.x1 = selarea.y1 = selarea.x2 = selarea.y2 = -1;

	prescan = 0;
      }

      if(fb_select_area(&selarea)) { /* area selected */
	ss.x = fbss.x + pixels2scanner(selarea.x1, fbss.xdpi);
	ss.y = fbss.y + pixels2scanner(selarea.y1, fbss.ydpi);
	ss.width = pixels2scanner(selarea.x2-selarea.x1+1, fbss.xdpi);
	ss.height = pixels2scanner(selarea.y2-selarea.y1+1, fbss.ydpi);
      }

      switch(selarea.key) {
	/* quit program */
      case 'q':
	quit = 1;
	break;

	  /* reset settings */
      case 'r':
	ss = set;
	prescan = 1;
	break;

	  /* set mark on settings */
      case 'm':
	set = ss;
	break;

	/* print settings  */
      case 't':
	fprintf(stderr,"Current settings:\n"
		"x-position: %f mm\n"
		"y-position: %f mm\n"
		"width: %f mm\n"
		"height: %f mm\n"
		"brightness: %f\n"
		"contrast: %f\n"
		"xdpi: %d\n"
		"ydpi: %d\n",
		scanner2mm(ss.x), scanner2mm(ss.y),
		scanner2mm(ss.width), scanner2mm(ss.height),
		ss.brightness, ss.contrast,
		ss.xdpi, ss.ydpi);

	if(ss.xdpi == ss.ydpi)
	  fprintf(stderr,"Command line: -d %d -b %.2f -c %.2f "
		  "-M -x %.2f -y %.2f -w %.2f -h %.2f\n",
		  ss.xdpi, ss.brightness, ss.contrast,
		  scanner2mm(ss.x), scanner2mm(ss.y),
		  scanner2mm(ss.width), scanner2mm(ss.height));
	else
	  fprintf(stderr,"Command line: --xdpi %d --ydpi %d -b %.2f -c %.2f "
		  "-M -x %.2f -y %.2f -w %.2f -h %.2f\n",
		  ss.xdpi, ss.ydpi, ss.brightness, ss.contrast,
		  scanner2mm(ss.x), scanner2mm(ss.y),
		  scanner2mm(ss.width), scanner2mm(ss.height));
	break;


	  /* do prescan to fb */
      case 'p':
	prescan = 1;
	break;

	/* do real scan to file */
      case 's':
	if(scan_to_file(scanfd, &ss) == -1)
	  quit = 1;
	break;

	/* Read new dpi values */
      case 'd':
	{
	  int new_dpi;
	  char new_dpi_text[16];

	  fb_go_tty();

	  fprintf(stderr,"New xdpi: [%d] ", ss.xdpi);
	  if(fgets(new_dpi_text, 15, stdin) == NULL) {
	    fprintf(stderr,"Couldn't set new dpi for some reason!\n");
	  } else {
	    /* only pressing return keeps the old value */
	    if(new_dpi_text[0] != '\n') { 
	      new_dpi = atoi(new_dpi_text);
	      if(new_dpi <= 0)
		fprintf(stderr,"New dpi zero or negative! That's not tolerable!\n");
	      else {
		ss.xdpi = new_dpi;
		if(auto_prescan)
		  prescan = 1;
	      }
	    }

	    if(verbose)
	      fprintf(stderr,"xdpi set to %d\n", ss.xdpi);
	  }

	  fprintf(stderr,"New ydpi: [%d] ", ss.ydpi);
	  if(fgets(new_dpi_text, 15, stdin) == NULL) {
	    fprintf(stderr,"Couldn't set new dpi for some reason!\n");
	  } else {
	    /* only pressing return keeps the old value */
	    if(new_dpi_text[0] != '\n') { 
	      new_dpi = atoi(new_dpi_text);
	      if(new_dpi <= 0)
		fprintf(stderr,"New dpi zero or negative! That's not tolerable!\n");
	      else {
		ss.ydpi = new_dpi;
		if(auto_prescan)
		  prescan = 1;
	      }
	    }

	    if(verbose)
	      fprintf(stderr,"ydpi set to %d\n", ss.ydpi);
	  }

	  fb_go_fb();
	}
      break;

	  /* Read new brightness value */
      case 'b':
	{
	  double new_bright;
	  char new_bright_text[16];

	  fb_go_tty();

	  fprintf(stderr,"New brightness (-1.0 - 1.0): [%f] ", ss.brightness);
	  if(fgets(new_bright_text, 15, stdin) == NULL) {
	    fprintf(stderr,"Couldn't set new brightness for some reason!\n");
	  } else {
	    /* only pressing return keeps the old value */
	    if(new_bright_text[0] != '\n') { 
	      new_bright = atof(new_bright_text);
	      if(new_bright < -1.0 || new_bright > 1.0)
		fprintf(stderr,"New brightness out of range! That's not tolerable!\n");
	      else {
		ss.brightness = new_bright;
		if(auto_prescan)
		  prescan = 1;
	      }
	    }

	    if(verbose)
	      fprintf(stderr,"brightness set to %f\n", ss.brightness);
	  }

	  fb_go_fb();
	}
      break;

	  /* Read new contrast value */
      case 'c':
	{
	  double new_cont;
	  char new_cont_text[16];

	  fb_go_tty();

	  fprintf(stderr,"New contrast (-1.0 - 1.0): [%f] ", ss.contrast);
	  if(fgets(new_cont_text, 15, stdin) == NULL) {
	    fprintf(stderr,"Couldn't set new contrast for some reason!\n");
	  } else {
	    /* only pressing return keeps the old value */
	    if(new_cont_text[0] != '\n') { 
	      new_cont = atof(new_cont_text);
	      if(new_cont < -1.0 || new_cont > 1.0)
		fprintf(stderr,"New contrast out of range! That's not tolerable!\n");
	      else {
		ss.contrast = new_cont;
		if(auto_prescan)
		  prescan = 1;
	      }
	    }

	    if(verbose)
	      fprintf(stderr,"contrast set to %f\n", ss.contrast);
	  }

	  fb_go_fb();
	}
      break;

	  /* Print help on interactive commands */
      case 'h':
	fb_go_tty();
	fprintf(stderr,"Interactive commands:\n"
		"    q - Quit program\n"
		"    r - Reset scanning area and prescan again\n"
		"    m - Set mark for settings to where a reset sets\n"
		"    p - Do prescan and view\n"
		"    s - Do real scan to a file (specified with -o, or to stdout)\n"
		"    t - Print current settings, with coordinates in mm\n"
		"    d - Enter new dpi values, xdpi and ydpi separately\n"
		"    b - Enter new brightness value\n"
		"    c - Enter new contrast value\n"
		"    h - Print this text\n"
		"Shift - Hide select box while pressed\n");
	fprintf(stderr,"Press Enter to return to image\n");
	getchar();
	fb_go_fb();
	break;

      default:
      }
    }
    pnm_close(scan);
    fb_exit();
  } else { /* no frame buffer stuff, just scan! */
    scan_to_file(scanfd, &ss);
  }

  close(scanfd);

  return 0;
}

int scan_to_file(int scanfd, ScanSettings *ss)
{
  int outfd;

  if(!outfile || !strcmp(outfile,"-"))
    outfd = STDOUT_FILENO;
  else {
    outfd = open(outfile, O_CREAT | O_WRONLY | O_TRUNC, 00644);
    if(outfd == -1) {
      perror("open(outfile)");
      return -1;
    }
  }

  if(hp_scan_file(outfd, scanfd, ss) == -1) {
    fprintf(stderr,"hp_scan_file(): Couldn't scan!\n");
    if(outfd != STDOUT_FILENO)
      close(outfd);
    return -1;
  }

  if(outfd != STDOUT_FILENO)
    close(outfd);

  return 0;
}


extern char *optarg;
extern int optind, opterr, optopt;

int parse_options(int argc, char *argv[], ScanSettings *ss)
{
  int arg;
  double posfactor = 300.0 / 25.4;  /* default is mm */
  double tmp_x=0.0,tmp_y=0.0,tmp_w=0.0,tmp_h=0.0;

  struct option long_options[] = { { "dpi", required_argument, NULL, 'd' },
				   { "xdpi", required_argument, NULL, 'X' },
				   { "ydpi", required_argument, NULL, 'Y' },
				   { "in", no_argument, NULL, 'I' },
				   { "cm", no_argument, NULL, 'C' },
				   { "mm", no_argument, NULL, 'M' },
				   { "pixels", no_argument, NULL, 'P' },
				   { "x-position", required_argument, NULL, 'x' },
				   { "y-position", required_argument, NULL, 'y' },
				   { "width", required_argument, NULL, 'w' },
				   { "height", required_argument, NULL, 'h' },
				   { "brightness", required_argument, NULL, 'b' },
				   { "contrast", required_argument, NULL, 'c' },
				   { "format", required_argument, NULL, 'f' },
				   { "prescan", required_argument, NULL, 'p' },
				   { "out", required_argument, NULL, 'o' },
				   { "info", required_argument, NULL, 'i' },
				   { "no-fb", no_argument, NULL, 'n' },
				   { "auto-prescan", no_argument, NULL, 'a' },
				   { "verbose", no_argument, NULL, 'v' },
				   { "help", no_argument, NULL, 'H' },
				   { "version", no_argument, NULL, 'V' } };
  char *short_options = "d:ICMPx:y:w:h:b:c:f:p:o:i:nav";

  while((arg=getopt_long(argc, argv, short_options, long_options, NULL)) != EOF) {
    switch(arg) {
    case 'd':
      ss->xdpi = ss->ydpi = atoi(optarg);
      break;
    case 'X':
      ss->xdpi = atoi(optarg);
      break;
    case 'Y':
      ss->xdpi = atoi(optarg);
      break;
    case 'I':
      posfactor = 300.0;
      break;
    case 'C':
      posfactor = 300.0 / 2.54;
      break;
    case 'M':
      posfactor = 300.0 / 25.4;
      break;
    case 'P':
      posfactor = 0;
      break;
    case 'x':
      tmp_x = atof(optarg);
      break;
    case 'y':
      tmp_y = atof(optarg);
      break;
    case 'w':
      tmp_w = atof(optarg);
      break;
    case 'h':
      tmp_h = atof(optarg);
      break;
    case 'b':
      ss->brightness = atof(optarg);
      break;
    case 'c':
      ss->contrast = atof(optarg);
      break;
    case 'f':
      if(!strcmp(optarg,"thresholded") || !strcmp(optarg,"t"))
	ss->format = FORMAT_THRESHOLDED;
      else if(!strcmp(optarg,"dithered") || !strcmp(optarg,"d"))
	ss->format = FORMAT_DITHERED;
      else if(!strcmp(optarg,"grayscale") || !strcmp(optarg,"g") ||
	      !strcmp(optarg,"greyscale"))
	ss->format = FORMAT_GRAYSCALE;
      else if(!strcmp(optarg,"color") || !strcmp(optarg,"c") ||
	      !strcmp(optarg,"colour"))
	ss->format = FORMAT_TRUECOLOR;
      else
	fprintf(stderr,"%s: Unknown format `%s' ignored!\n",argv[0],optarg);
      break;
    case 'p':
      if(!strcmp(optarg,"thresholded") || !strcmp(optarg,"t"))
	prescan_format = FORMAT_THRESHOLDED;
      else if(!strcmp(optarg,"dithered") || !strcmp(optarg,"d"))
	prescan_format = FORMAT_DITHERED;
      else if(!strcmp(optarg,"grayscale") || !strcmp(optarg,"g") ||
	      !strcmp(optarg,"greyscale"))
	prescan_format = FORMAT_GRAYSCALE;
      else if(!strcmp(optarg,"color") || !strcmp(optarg,"c") ||
	      !strcmp(optarg,"colour"))
	prescan_format = FORMAT_TRUECOLOR;
      else
	fprintf(stderr,"%s: Unknown format `%s' ignored!\n",argv[0],optarg);
      break;
    case 'o':
      outfile = optarg;
      break; 
    case 'i':
      infotext = optarg;
      break;
    case 'n':
      no_fb = 1;
      break;
    case 'a':
      auto_prescan = 1;
      break;
    case 'v':
      verbose = 1;
      break;
    case 'H':
      fprintf(stderr,"Usage: %s [OPTION(S)]\n"
	      "\n"
	      "valid options are:\n"
	      "-d, --dpi n           sets x & y scan resolution in dots per inch\n"
	      "    --xdpi n          sets x scan resolution only (default 50)\n"
	      "    --ydpi n          sets y scan resolution only (default 50)\n"
	      "-I, --in              specify scan region in inches\n"
	      "-C, --cm              specify scan region in centimeters\n"
	      "-M, --mm              specify scan region in millimeters (default)\n"
	      "-P, --pixels          specify scan region in pixels (see note!)\n"
	      "-x, --x-position f    specify upper left x of scan region (default 0.0)\n"
	      "-y, --y-position f    specify upper left y of scan region (default 0.0)\n"
	      "-w, --width f         specify width of scan region (default 210mm)\n"
	      "-h, --height f        specify height of scan region (default 297mm)\n"
	      "-b, --brightness f    specify brightness adjustment (-1.0 to 1.0, default 0.0)\n"
	      "-c, --contrast f      specify contrast adjustment (-1.0 to 1.0, default 0.0)\n"
	      "-s, --device <name>   scsi generic device of scanner (default autoprobe)\n"
	      "-o, --out <name>      output file (default is stdout)\n"
	      "-i, --info <text>     set own info text in PNM file\n"
	      "-f, --format <mode>   modes: thresholded, dithered, greyscale, colour (default)\n"
	      "-p, --prescan <mode>  mode when prescanning, see -f, greyscale is default\n"
	      "-n, --no-fb           don't use frame buffer, only scan to file\n"
	      "-a, --auto-prescan    always prescan after each new change of settings\n"
	      "-v, --verbose         print informational messages\n"
	      "    --version         print version and exit\n"
	      "    --help            print this text and exit\n"
	      "\n"
	      "n represents an integer number, f a float.\n"
	      "Options -I, -C, -M and -P sets the measure type for all relevant\n"
	      "  values on the commandline, no matter what order the options are in.\n"
	      "Option -P (--pixels) depends on the dpi values used!\n"
	      "  The pixel values typed in does not have to be the size of\n"
	      "  the scanned picture!\n"
              "Later options always override previous ones.\n"
              ,argv[0]);
      return 1;
      break;
    case 'V':
      fprintf(stderr,PROGRAM_NAME " version %d.%d.%d%s\n", 
	      PROGRAM_VERSION_MAJOR, PROGRAM_VERSION_MINOR,
	      PROGRAM_VERSION_SUBMINOR, PROGRAM_VERSION_EXTENSION);
      return 1;
      break;
    default:
    }
  }

  if(posfactor == 0) { /* pixels used for width/height */
    if(tmp_x) ss->x = pixels2scanner(tmp_x, ss->xdpi);
    if(tmp_y) ss->y = pixels2scanner(tmp_y, ss->ydpi);
    if(tmp_w) ss->width = pixels2scanner(tmp_w, ss->xdpi);
    if(tmp_h) ss->height = pixels2scanner(tmp_h, ss->ydpi);
  } else {
    if(tmp_x) ss->x = tmp_x * posfactor;
    if(tmp_y) ss->y = tmp_y * posfactor;
    if(tmp_w) ss->width = tmp_w * posfactor;
    if(tmp_h) ss->height = tmp_h * posfactor;
  }

  return 0;
}
