/* pnm.c - Copyright (C) Tomas Berndtsson, tomas@nocrew.org
 *
 * Some simple functions to handle pnm files and
 * to convert them into some viewable format for 
 * the frame buffer.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "fb.h"
#include "pnm.h"

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))

static int pnm_is_space(char ch)
{
  static int comment=0;

  if(comment) {
    if(ch != '\n')
      return 1;
    else {
      comment = 0;
      return 1;
    }
  } else {
    if(isspace(ch))
      return 1;
    if(ch == '#') {
      comment = 1;
      return 1;
    }
  }

  return 0;
}

static int pnm_skip_spaces(FILE *file)
{
  char ch;

  ch = fgetc(file);
  while(ch != EOF && pnm_is_space(ch))
    ch = fgetc(file);

  if(ch == EOF)
    return EOF;

  fseek(file, -1, SEEK_CUR);

  return 0;
}

PNMPic *pnm_load(char *filename)
{
  FILE *file;
  PNMPic *pnm;
  int mode,xres,yres,cols;
  char ch;
  
  file = fopen(filename, "r");
  if(file == NULL) {
    fprintf(stderr,"pnm_load(): Couldn't open file!\n");
    return NULL;
  }

  pnm = (PNMPic *)malloc(sizeof(PNMPic));
  if(pnm == NULL) {
    fprintf(stderr,"pnm_load(): Not enough memory for PNMPic struct!\n");
    fclose(file);
    return NULL;
  }

  ch = fgetc(file);
  if(ch == 'P') {
    ch = fgetc(file);
    mode = ch - 0x30;
  } else {
    fprintf(stderr,"pnm_load(): Couldn't find initializing 'P'!\n");
    fclose(file);
    free(pnm);
    return NULL;
  }
  if(mode < 4) {
    fprintf(stderr,"pnm_load(): Can only handle rawbits PNMs right now!\n");
    fclose(file);
    free(pnm);
    return NULL;
  }

  if(!pnm_skip_spaces(file)) {
    fscanf(file, "%d", &xres);
  } else {
    fprintf(stderr,"pnm_load(): Something went wrong around xres!\n");
    fclose(file);
    free(pnm);
    return NULL;
  }

  if(!pnm_skip_spaces(file)) {
    fscanf(file, "%d", &yres);
  } else {
    fprintf(stderr,"pnm_load(): Something went wrong around yres!\n");
    fclose(file);
    free(pnm);
    return NULL;
  }

  if(mode > 4) {
    if(!pnm_skip_spaces(file)) {
      fscanf(file, "%d", &cols);
    } else {
      fprintf(stderr,"pnm_load(): Something went wrong around cols!\n");
      fclose(file);
      free(pnm);
      return NULL;
    }
  } else {
    cols = 1;
  }

  /* remove last newline. maybe check something here? */
  ch = fgetc(file);

  pnm->mode = mode;
  pnm->xres = xres;
  pnm->yres = yres;
  pnm->max_color = cols;
  if(mode == 4)
    pnm->length = (xres*yres)/8;
  else if(mode == 5)
    pnm->length = xres*yres;
  else if(mode == 6)
    pnm->length = xres*yres*3;

#ifdef DEBUG
  fprintf(stderr,"pnm->mode: %d\n",pnm->mode);
  fprintf(stderr,"pnm->xres: %d\n",pnm->xres);
  fprintf(stderr,"pnm->yres: %d\n",pnm->yres);
  fprintf(stderr,"pnm->max_color: %d\n",pnm->max_color);
  fprintf(stderr,"pnm->length: %d\n",pnm->length);
#endif /* DEBUG */

  pnm->buf = (char *)malloc(pnm->length);
  if(pnm->buf == NULL) {
    fprintf(stderr,"pnm_load(): Not enough memory for pnm data!\n");
    fclose(file);
    free(pnm);
    return NULL;
  }

  if(fread(pnm->buf, 1, pnm->length, file) != pnm->length) {
    fprintf(stderr,"pnm_load(): Didn't read all of the file for some reason!\n");
    fclose(file);
    free(pnm->buf);
    free(pnm);
    return NULL;
  }

  fclose(file);
  return pnm;
}

void pnm_close(PNMPic *pic)
{
  if(pic != NULL) {
    if(pic->buf != NULL)
      free(pic->buf);
    free(pic);
  }
}

int pnm_draw(FB *fb, PNMPic *pic)
{
  int x,y,bit,shift,nr_cols,col;
  int *red,*green,*blue;
  int xres,yres,xdiff,ydiff;
  unsigned char *data;

  /* can only handle rawbits pnm's at the moment! */
  if(pic->mode < 4)
    return -1;

  nr_cols = 1<<fb->vinf.bits_per_pixel;
  if((red   = malloc(nr_cols*sizeof(unsigned int))) == NULL)
    return -1;
  if((green = malloc(nr_cols*sizeof(unsigned int))) == NULL)
    return -1;
  if((blue  = malloc(nr_cols*sizeof(unsigned int))) == NULL)
    return -1;

  for(col=0 ; col<nr_cols ; col++) {
    red[col]=green[col]=blue[col] = col ? ((65536*(col+1))/nr_cols)-1 : 0;
  }
  fb_setpalette(red, green, blue);

  shift = (pic->max_color+1) / nr_cols;

  xres = min(fb->vinf.xres, pic->xres);
  yres = min(fb->vinf.yres, pic->yres);
  xdiff = pic->xres - xres;
  ydiff = pic->yres - yres;
  
  data = pic->buf;

  if(pic->mode == 4) {
    /* This can be made much faster using FBbitblt() 
       but I haven't bothered yet */
    for(y=0 ; y<yres ; y++) {
      for(x=0 ; x<xres ; x+=8) {
	for(bit=0 ; (bit<8) && (x+bit<xres) ; bit++) {
	  col = (((*data) << bit) & 0x80) ? 0 : nr_cols-1;
	  fb_putpixel(x+bit, y, col);
	}
	data++;
      }
      data += xdiff/8;
    }
  } else if(pic->mode == 5) {
    for(y=0 ; y<yres ; y++) {
      for(x=0 ; x<xres ; x++) {
	col = (*data++) / shift;
	fb_putpixel(x, y, col);
      }
      data += xdiff;
    }
  } else if(pic->mode == 6) {
    for(y=0 ; y<yres ; y++) {
      for(x=0 ; x<xres ; x++) {
	/* make approximate greyscale from truecolour values */
	col = (data[0] + data[1] + data[2]) / (shift*3);
	fb_putpixel(x, y, col);
	data += 3;
      }
      data += xdiff*3;
    }
  } else
    return -1;

  return 0;
}

